/*
 * File:        keyboard_gloves.c
 * Author:      Vitchyr Pong, Gulnar Mirza
 * Adapted from:
 *              main.c by
 * Author:      Bruce Land
 * Target PIC:  PIC32MX250F128B
 */

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config.h"
// threading library
#include "pt_cornell_1_2.h"

////////////////////////////////////
// graphics libraries
#include "tft_master.h"
#include "tft_gfx.h"
// need for rand function
#include <stdlib.h>
#include <math.h>
////////////////////////////////////


/* Demo code for interfacing TFT (ILI9340 controller) to PIC32
 * The library has been modified from a similar Adafruit library
 */
// Adafruit data:
/***************************************************
  This is an example sketch for the Adafruit 2.2" SPI display.
  This library works with the Adafruit 2.2" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/1480

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/
// Functions
// PORT B
#define EnablePullDownB(bits) CNPUBCLR=bits; CNPDBSET=bits;
#define DisablePullDownB(bits) CNPDBCLR=bits;
#define EnablePullUpB(bits) CNPDBCLR=bits; CNPUBSET=bits;
#define DisablePullUpB(bits) CNPUBCLR=bits;
//PORT A
#define EnablePullDownA(bits) CNPUACLR=bits; CNPDASET=bits;
#define DisablePullDownA(bits) CNPDACLR=bits;
#define EnablePullUpA(bits) CNPDACLR=bits; CNPUASET=bits;
#define DisablePullUpA(bits) CNPUACLR=bits;

// Constants
#define	SYS_FREQ 40000000 // 40 MHz

// State machine constants
#define NO_PUSH 0
#define MAYBE_PUSH 1
#define PUSHED 2
#define MAYBE_NO_PUSH 3

// === ADC Macros ====================================================
// define setup parameters for OpenADC10
// Turn module on | ouput in integer | trigger mode auto | enable autosample
// ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
// ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
// ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
#define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON

// define setup parameters for OpenADC10
// ADC ref external  | disable offset test | disable scan mode | do 1 sample | use single buf | alternate mode off
#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF

// Define setup parameters for OpenADC10
// use peripherial bus clock | set sample time | set ADC clock divider
// ADC_CONV_CLK_Tcy2 means divide CLK_PB by 2 (max speed)
// ADC_SAMPLE_TIME_5 seems to work with a source resistance < 1kohm
#define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_5 | ADC_CONV_CLK_Tcy2 //ADC_SAMPLE_TIME_15| ADC_CONV_CLK_Tcy2

// define setup parameters for OpenADC10
// set AN11 and  as analog inputs
#define PARAM4	ENABLE_AN11_ANA // pin 24

// define setup parameters for OpenADC10
// do not assign channels to scan
#define PARAM5	SKIP_SCAN_ALL

// AVOID RPB 6 & 10
#define ENCODER_BPORT_BITS (BIT_4 | BIT_5 | BIT_2)
#define MUX_SELECT_FLEX_BPORT_BITS (BIT_7 | BIT_8 | BIT_9)

#define MUX_SELECT_BUTTON_BPORT_BITS (BIT_0 | BIT_1 | BIT_2)
#define MUX_BUTTON_BPORT_READ_BIT BIT_15

#define LEFT_INDEX_APORT_BIT BIT_3
#define RIGHT_INDEX_APORT_BIT BIT_2

#define SPACE_BPORT_BIT BIT_3

#define INIT_MAX_THRESH (1024/3*2)
#define INIT_MIN_THRESH (1024/3)

#define N_COLS 10
#define N_ROWS 3
#define N_SAMPLES 1

#define CALIBRATE 1

// User interface variables
char buffer[60];
static char cmd[16];
static float value;
static int flex_val;

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_glove, pt_computer_interface, pt_debug;

// Pt structs for UART
static struct pt pt_input, pt_output, pt_DMA_output;

// PT for helper methods
static struct pt pt_handle_push_button;

static int encoder_input;
static int mux_bits;

// Push button logic variables
static int push_state, pressed;

// Key pressing variables
static char key_pressed;
static short new_key_pressed; // boolean flag

// Calibration variables
static short calibrating; // flag for whether or not we are calibrating
static char char_to_calibrate; // what character we are calibrating
// what column and row are we calibrating? Top left is (0, 0)
static int calib_row, calib_col;
// what column and row will we calibrate next? Top left is (0, 0)
static int next_calib_row, next_calib_col;
// How many times we've sampled a given column and row
static int calib_sample;
static int next_calib_sample;

// Convert the output of an encoder (3 bit) to a column number.
int encoder_output_to_col_num(int encoder_val) {
    // TODO: update these values based on the actual encoder implementation.
    if (encoder_val == BIT_4) {
        return 0;
    }
    if (encoder_val == BIT_5) {
        return 1;
    }
    if (encoder_val == (BIT_4 | BIT_5)) {
        return 2;
    }
    return -1; // return error for now
}

/* Convert the column number to the mux bits */
int col_num_to_mux_bits(int col_num) {  
    switch (col_num) {
        case 9: return 0;
        case 8: return BIT_7;
        case 7: return BIT_8;
        case 6: return BIT_8 | BIT_7; // col 3 & 4 are both left index finger
        case 5: return BIT_8 | BIT_7;
        case 4: return BIT_9; // col 5 & 6 are both right index finger
        case 3: return BIT_9;
        case 2: return BIT_9 | BIT_7;
        case 1: return BIT_9 | BIT_8;
        case 0: return BIT_9 | BIT_8 | BIT_7;
        default: return 0;
    }
}
// === Glove Thread =================================================
/* Maps a port number and row to a character */
/* Don't ask why, I had to flip everything twice.
static char row_and_col_to_char[N_ROWS][N_COLS] = {
    {'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p'},
    {'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';'},
    {'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/'}
};
*/

static char row_and_col_to_char[N_ROWS][N_COLS] = {
    {'p', 'o', 'i', 'u', 'y', 't', 'r', 'e', 'w', 'q'},
    {';', 'l', 'k', 'j', 'h', 'g', 'f', 'd', 's', 'a'},
    {'/', '.', ',', 'm', 'n', 'b', 'v', 'c', 'x', 'z'}
};

/* The measured flex values for different key */
static int row_and_col_to_sample_flex_val[N_ROWS][N_COLS][N_SAMPLES];

/* row_and_col_to_thresholds[r][c] gives the threshold for the flex value
 * between rows r and r+1 at column c of the keyboard. */
/*
static int row_and_col_to_thresholds[N_ROWS-1][N_COLS] = {
    {481, 851, 455, 762, 757, 323, 326, 607, 650, 261},
    {907, 1023, 877, 968, 999, 673, 706, 892, 950, 740}
};
*/
static int row_and_col_to_thresholds[N_ROWS-1][N_COLS] = {
    {69, 688, 172, 496, 537, 263, 266, 508, 344, 28},
    {201, 847, 287, 685, 733, 546, 546, 653, 730, 254}
};

/* Given a column and flex value, return the key pressed*/
static char get_key_pressed(int c, int flex_v) {
    int keyboard_row = 0;
    int r;
    for (r = 1; r < N_ROWS; r++) {
        // minus one since row_thresholds gives the values _between_ the rows
        if (flex_v > row_and_col_to_thresholds[r-1][c]) {
            keyboard_row = r;
        }
    }
    return row_and_col_to_char[keyboard_row][c];
}

// Calibrate a key given its flex value.
void calibrate_key(int val) {
    row_and_col_to_sample_flex_val[calib_row][calib_col][calib_sample] = val;
}

/* Given that a column of the keyboard is pressed, find out what the key is.
 *
 * If [is_space_pushed] is true, then the col_pressed is ignore.
 **/
void handle_pushed_button(int col_pressed, int is_space_pushed) {
    flex_val = ReadADC10(0);
    
    if (calibrating) {
        calibrate_key(flex_val);
        key_pressed = char_to_calibrate;
    } else {
        key_pressed = get_key_pressed(col_pressed, flex_val);
    }
    new_key_pressed = 1;
}

// Set the mux to select for column c.
void set_col_bits(int c) {
    mPORTBClearBits(MUX_SELECT_BUTTON_BPORT_BITS);
    int mux_b_bits = 0;
    switch (c) {
        case 9: break;
        case 8: mux_b_bits = BIT_0; break;
        case 7: mux_b_bits = BIT_1; break;
        case 6: mux_b_bits = BIT_0 | BIT_1; break; // col 3 & 4 are both left index finger
        case 5: mux_b_bits = BIT_0 | BIT_1; break;
        case 4: mux_b_bits = BIT_2; break; // col 5 & 6 are both left index finger
        case 3: mux_b_bits = BIT_2; break;
        case 2: mux_b_bits = BIT_0 | BIT_2; break;
        case 1: mux_b_bits = BIT_1 | BIT_2; break;
        case 0: mux_b_bits = BIT_0 | BIT_1 | BIT_2; break;
    }
    mPORTBSetBits(mux_b_bits);
}

static int col_pressed = -1;
static int c;
// out = finger is in the middle rows
static int is_left_out, is_right_out;
static int space_was_pushed;
// === Glove Thread =================================================
// Read the input from the glove
static PT_THREAD (protothread_glove(struct pt *pt))
{
    PT_BEGIN(pt);
        while(1) {
            col_pressed = -1;
            space_was_pushed = 0;
            if (mPORTBReadBits(SPACE_BPORT_BIT) != 0) {
                space_was_pushed = 1;
            } else {
                for (c = 0; c < N_COLS; c++) {
                    set_col_bits(c);
                    PT_YIELD_TIME_msec(3);
                    // If index sensors are apart, then there's only the
                    // internal pull-down resistor.
                    is_left_out = mPORTAReadBits(LEFT_INDEX_APORT_BIT) == 0;
                    is_right_out = mPORTAReadBits(RIGHT_INDEX_APORT_BIT) == 0;
                    if (mPORTBReadBits(MUX_BUTTON_BPORT_READ_BIT) != 0) {
                        col_pressed = c;
                        break;
                    }
                    mPORTBClearBits(MUX_SELECT_BUTTON_BPORT_BITS);

                    // skip 4 and 6 since the left index finger counts for columns 3
                    // and 4. The right index finger counts for columns 5 and 6.
                    if (c == 3 || c == 5) {
                        c++;
                    }
                }
            }
            
            if (col_pressed == 3 &&  is_left_out) {
                col_pressed = 4;
            }
            if (col_pressed == 5 && !is_right_out) {
                col_pressed = 6;
            }

            pressed = space_was_pushed || col_pressed != -1;
            switch (push_state) {
                case NO_PUSH:
                    if (pressed) {
                        push_state=MAYBE_PUSH;
                    } else {
                        push_state=NO_PUSH;
                    }
                    break;

                case MAYBE_PUSH:
                    if (pressed) {
                        push_state=PUSHED;
                        if (space_was_pushed && !calibrating) { // Ignore space when calibrating
                            key_pressed = ' ';
                            new_key_pressed = 1;
                        } else {
                            mux_bits = col_num_to_mux_bits(col_pressed);
                            mPORTBClearBits(MUX_SELECT_FLEX_BPORT_BITS);
                            mPORTBSetBits(mux_bits);
                            PT_YIELD_TIME_msec(10);
                            handle_pushed_button(col_pressed, space_was_pushed);
                        }
                    } else {
                        push_state=NO_PUSH;
                    }
                    break;

               case PUSHED:  
                  if (pressed) {
                      push_state=PUSHED; 
                  } else {
                      push_state=MAYBE_NO_PUSH;    
                  }
                  break;

               case MAYBE_NO_PUSH:
                  if (pressed) {
                      push_state=PUSHED; 
                  } else {
                      push_state=NO_PUSH;    
                  }
                  break;
            }
            PT_YIELD_TIME_msec(1);
            // NEVER exit while
        } // END WHILE(1)
    PT_END(pt);
} // Glove thread


// Get the sample value of a given row and col.
int get_sample_value(int r, int c) {
    int sum = 0;
    int s;
    for (s = 0; s < N_SAMPLES; s++) {
        sum += row_and_col_to_sample_flex_val[r][c][s];
    }
    return sum / N_SAMPLES;
}

/* Calibrate the thresholds based on the sample flex values.
 *
 * We have a flex value for each row. The idea of this calibration is to choose
 * thresholds between each row. We do this by taking the average of consecutive
 * rows and settings those as the thresholds.
 */
void finish_calibration() {
    int r, c;
    for (c = 0; c < N_COLS; c++) {
        int last_val = get_sample_value(0, c);
        for (r = 1; r < N_ROWS; r++) {
            int val = get_sample_value(r, c);
            row_and_col_to_thresholds[r-1][c] = (last_val + val) / 2;
            last_val = val;
        }
    }
}

// === User Interface Thread =================================================
/* Interface with computer via UART. The format of the string should be either:
 * 
 *      "k <character>"
 *          user pressed <character> key
 * 
 *      "c <character>"
 *          prompt user to press <character> to calibrate.
 * 
 *      "d <data>"
 *          <data> is info, hopefully useful for debugging. <data> should be a
 *          space-separated list of "<key>:<value>" strings. Note that there is
 *          no space between the colon and the key nor the value.
 *
 *      "s"
 *          Indicates that a space was pressed.
 */
static PT_THREAD (protothread_computer_interface(struct pt *pt))
{
    PT_BEGIN(pt);
        while(1) {
            if (calibrating) {
                for (calib_sample = 0; calib_sample < N_SAMPLES; calib_sample++) {
                    for (calib_col = 0; calib_col < N_COLS; calib_col++) {
                        for (calib_row = 0; calib_row < N_ROWS; calib_row++) {
                            char_to_calibrate = row_and_col_to_char[calib_row][calib_col];
                            sprintf(PT_send_buffer,"c %c\n", char_to_calibrate);
                            // send character to computer by spawning a print thread
                            PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                            new_key_pressed = 0;
                            PT_YIELD_UNTIL(pt, new_key_pressed);
                        }
                    }
                }
                finish_calibration();
                calibrating = 0;
            } else {
                if (key_pressed == ' ') {
                    sprintf(PT_send_buffer,"s\n");
                } else {
                    sprintf(PT_send_buffer,"k %c\n", key_pressed);
                }
                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));

                // also send some debugging data
                sprintf(PT_send_buffer,"d val:%d\n", flex_val);
                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
            }
            new_key_pressed = 0;
            PT_YIELD_UNTIL(pt, new_key_pressed);
            // NEVER exit while
        } // END WHILE(1)
    PT_END(pt);
} // User Interface thread

// === Debug Thread =================================================
static int c_print = 0;
static int print_thresh = 1;
static int done_printing = 0;
// Send information, mostly for debugging.
static PT_THREAD (protothread_debug(struct pt *pt))
{
    PT_BEGIN(pt);
        while(1) {
            if (!calibrating && !done_printing) {
                if (print_thresh) {
                    for (c_print=0; c_print < N_COLS; c_print++) {
                        sprintf(PT_send_buffer,"d 0-1:%d 1-2:%d\n", row_and_col_to_thresholds[0][c_print], row_and_col_to_thresholds[1][c_print]);
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                        PT_YIELD_TIME_msec(250);
                    }
                    print_thresh = 0;
                } else {
                    for (c_print=0; c_print < N_COLS; c_print++) {
                        sprintf(PT_send_buffer,"d c:%d r0:%d r1:%d r2:%d\n",
                            c_print,
                            row_and_col_to_sample_flex_val[0][c_print][0],
                            row_and_col_to_sample_flex_val[1][c_print][0],
                            row_and_col_to_sample_flex_val[2][c_print][0]);
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                        PT_YIELD_TIME_msec(250);
                    }
                    done_printing = 1;
                }
            }
            
            PT_YIELD_TIME_msec(250);
            // NEVER exit while
        } // END WHILE(1)
    PT_END(pt);
} // Debug thread

// === Main  ======================================================
void main(void) {
    //SYSTEMConfigPerformance(PBCLK);

    ANSELA = 0; ANSELB = 0; 

    // === config threads ==========
    // turns OFF UART support and debugger pin, unless defines are set
    PT_setup();

    // === setup system wide interrupts  ========
    INTEnableSystemMultiVectoredInt();

    // === init the threads ===================================
    PT_INIT(&pt_computer_interface);
    PT_INIT(&pt_glove);
    PT_INIT(&pt_debug);
    
    // == the ADC ================================
    // configure and enable the ADC for the flex sensor value
    // PIN 24
    CloseADC10();	// ensure the ADC is off before setting the configuration

    // use ground as neg ref for A | use AN11 for input A (pin 24) 
    // configure to sample AN11 
    SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN11 ); // configure to sample AN4 
    OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using the parameters defined above

    EnableADC10(); // Enable the ADC
    
    // == init sensors ================================
    DisablePullDownA(0xff);
    DisablePullDownB(0xff);
    
    // Flex sensors mux control
    mPORTBSetPinsDigitalOut(MUX_SELECT_FLEX_BPORT_BITS);
    mPORTBClearBits(MUX_SELECT_FLEX_BPORT_BITS);

    // Button push mux control
    mPORTBSetPinsDigitalOut(MUX_SELECT_BUTTON_BPORT_BITS);
    mPORTBClearBits(MUX_SELECT_BUTTON_BPORT_BITS);
    // Enable the button push sensor
    mPORTBSetPinsDigitalIn(MUX_BUTTON_BPORT_READ_BIT);

    // Enable index finger sensors
    EnablePullDownA(LEFT_INDEX_APORT_BIT | RIGHT_INDEX_APORT_BIT);
    mPORTASetPinsDigitalIn(LEFT_INDEX_APORT_BIT | RIGHT_INDEX_APORT_BIT);
    
    // Enable space bar button sensor
    EnablePullDownB(SPACE_BPORT_BIT);
    mPORTBSetPinsDigitalIn(SPACE_BPORT_BIT);
    
    // Initialize default values to be safe/explicit
    new_key_pressed = 0;
    calibrating = CALIBRATE;
    calib_row = 0;
    calib_col = 0;
    next_calib_row = 0;
    next_calib_col = 0;
    
    // round-robin scheduler for threads
    while (1){
        PT_SCHEDULE(protothread_computer_interface(&pt_computer_interface));
        PT_SCHEDULE(protothread_glove(&pt_glove));
        PT_SCHEDULE(protothread_debug(&pt_debug));
    }
} // main

// === end  ======================================================

